from typing import Optional, Union

import numpy as np
import pandas as pd
from caveclient import CAVEclient
from joblib import Parallel, delayed
from requests.exceptions import HTTPError
from tqdm.auto import tqdm
from tqdm_joblib import tqdm_joblib

from .constants import TIMESTAMP_DELTA
from .replay import resolve_edit
from .types import Integer


def _sort_edgelist(edgelist: np.ndarray) -> np.ndarray:
    return np.unique(np.sort(edgelist, axis=1), axis=0)


def _get_level2_nodes_edges(
    root_id: Integer, client: CAVEclient, bounds: Optional[np.ndarray] = None
) -> tuple[np.ndarray, np.ndarray]:
    try:
        edgelist = client.chunkedgraph.level2_chunk_graph(root_id, bounds=bounds)
        nodelist = set()
        for edge in edgelist:
            for node in edge:
                nodelist.add(node)
        nodelist = list(nodelist)
    except HTTPError:
        # REF: https://github.com/seung-lab/PyChunkedGraph/issues/404
        nodelist = client.chunkedgraph.get_leaves(root_id, stop_layer=2)
        if len(nodelist) != 1:
            raise HTTPError(
                f"HTTPError: level 2 chunk graph not found for root_id: {root_id}"
            )
        else:
            edgelist = np.empty((0, 2), dtype=int)

    if len(edgelist) == 0:
        edgelist = np.empty((0, 2), dtype=int)
    else:
        edgelist = np.array(edgelist, dtype=int)

    edgelist = _sort_edgelist(edgelist)

    nodelist = np.array(nodelist, dtype=int)
    nodelist = np.unique(nodelist)

    return nodelist, edgelist


def get_node_aliases(
    supervoxel_id, client, stop_layer=2, return_as="list"
) -> Union[list, pd.DataFrame]:
    """For a given supervoxel, get the node that it was part of at `stop_layer` for
    each timestamp.
    """
    current_ts = client.timestamp

    node_id = client.chunkedgraph.get_roots(
        supervoxel_id, stop_layer=stop_layer, timestamp=current_ts
    )[0]
    oldest_ts = client.chunkedgraph.get_oldest_timestamp()

    node_info = []
    while current_ts > oldest_ts:
        created_ts = client.chunkedgraph.get_root_timestamps(node_id)[0]
        node_info.append(
            {
                "node_id": node_id,
                "start_valid_ts": created_ts,
                "end_valid_ts": current_ts,
            }
        )
        current_ts = created_ts - TIMESTAMP_DELTA
        node_id = client.chunkedgraph.get_roots(
            supervoxel_id, stop_layer=stop_layer, timestamp=current_ts
        )[0]

    node_info = pd.DataFrame(node_info).set_index("node_id")
    if return_as == "list":
        return node_info.index.to_list()
    elif return_as == "df":
        return node_info


# a version of the above that used the already computed edits to do the tracking

# current_timestamp = client.timestamp
# supervoxel_id = nuc_supervoxel_id

# level2_id = client.chunkedgraph.get_roots(supervoxel_id, stop_layer=2)[0]

# operation_id_added = None

# level2_id_info = []

# for operation_id, delta in tqdm(list(networkdeltas.items())[::-1], disable=True):
#     if level2_id in delta.added_nodes:
#         operation_id_added = operation_id
#         operation_added_ts = client.chunkedgraph.get_operation_details(
#             [operation_id_added]
#         )[str(operation_id_added)]["timestamp"]
#         operation_added_ts = datetime.fromisoformat(operation_added_ts)

#         level2_id_info.append(
#             {
#                 "level2_id": level2_id,
#                 "operation_id_added": operation_id_added,
#                 "start_valid_ts": operation_added_ts,
#                 "end_valid_ts": current_timestamp,
#             }
#         )

#         # get the new ID and continue to search backwards
#         pre_operation_added_ts = operation_added_ts - TIMESTAMP_DELTA
#         level2_id = client.chunkedgraph.get_roots(
#             nuc_supervoxel_id, stop_layer=2, timestamp=pre_operation_added_ts
#         )[0]
#         current_timestamp = pre_operation_added_ts


# level2_id_info.append(
#     {
#         "level2_id": level2_id,
#         "operation_id_added": None,
#         "start_valid_ts": None,
#         "end_valid_ts": current_timestamp,
#     }
# )
# pd.DataFrame(level2_id_info)


def get_component_masks(components: list[set]):
    """From a list of components, get a node by component boolean DataFrame of masks."""
    used_l2_nodes = np.unique(np.concatenate([list(c) for c in components]))
    l2_masks = pd.DataFrame(
        index=used_l2_nodes,
        data=np.zeros((len(used_l2_nodes), len(components)), dtype=bool),
    )
    for i, component in enumerate(components):
        l2_masks.loc[list(component), i] = True

    return l2_masks


def get_nucleus_supervoxel(root_id, client):
    """Get the supervoxel corresponding to the nucleus of a neuron by looking it up
    in the soma table."""
    nuc_table = client.info.get_datastack_info()["soma_table"]
    nuc_info = client.materialize.query_table(
        nuc_table, filter_equal_dict=dict(pt_root_id=root_id), log_warning=False
    )
    if len(nuc_info) == 0:
        raise ValueError(f"No nucleus found for root_id: {root_id}")
    elif len(nuc_info) > 1:
        raise ValueError(f"Multiple nuclei found for root_id: {root_id}")
    else:
        nuc_supervoxel_id = nuc_info["pt_supervoxel_id"].values[0]
    return nuc_supervoxel_id


def get_nodes_aliases(
    supervoxel_ids, client, stop_layer=2, verbose=True, return_as="list"
):
    """For a list of supervoxels, get all of the nodes at `stop_layer` that they were
    part of across time."""

    if not isinstance(supervoxel_ids, list):
        supervoxel_ids = list(supervoxel_ids)

    # for all supervoxels, look up their level2 node now
    current_ts = client.timestamp
    chunk_size = 50_000
    supervoxel_id_chunks = np.array_split(
        supervoxel_ids, len(supervoxel_ids) // chunk_size
    )
    node_ids_by_chunk = []
    for supervoxel_id_chunk in supervoxel_id_chunks:
        node_ids = client.chunkedgraph.get_roots(
            supervoxel_id_chunk, stop_layer=stop_layer, timestamp=current_ts
        )
        node_ids_by_chunk.append(node_ids)
    node_ids = np.concatenate(node_ids_by_chunk)
    node_to_supervoxel = pd.Series(index=node_ids, data=supervoxel_ids)
    node_to_supervoxel.index.name = "node_id"
    node_to_supervoxel.name = "supervoxel_id"

    # find out when they were made
    timestamps = client.chunkedgraph.get_root_timestamps(node_ids)
    node_timestamps = pd.Series(data=timestamps, index=node_ids)
    node_timestamps.index.name = "node_id"
    node_timestamps.name = "timestamp"

    # anything that was created at before oldest_ts we can ignore
    oldest_ts = client.chunkedgraph.get_oldest_timestamp()

    # everything else, we need to look up its history
    new_nodes = node_timestamps[node_timestamps > oldest_ts].index
    supervoxels_to_lookup = node_to_supervoxel.loc[new_nodes]

    # TODO this could be a bit faster if we wrote a smarter implementation since
    # some nodes may end up in the same history, don't think would be a huge speedup,
    # though
    with tqdm_joblib(total=len(supervoxels_to_lookup), disable=not verbose):
        historical_l2_ids = Parallel(n_jobs=-1)(
            delayed(lambda sv: get_node_aliases(sv, client))(sv)
            for sv in supervoxels_to_lookup
        )

    supervoxel_historical_l2_ids = {
        sv: l2s for sv, l2s in zip(supervoxels_to_lookup, historical_l2_ids)
    }
    # backfill the ones that we didnt have to go back in time for
    old_nodes = node_timestamps[node_timestamps <= oldest_ts].index
    old_supervoxels_to_nodes = (
        node_to_supervoxel.loc[old_nodes]
        .reset_index()
        .set_index("supervoxel_id")["node_id"]
    )
    old_supervoxels_to_nodes = old_supervoxels_to_nodes.apply(lambda x: [x])
    supervoxel_historical_l2_ids.update(old_supervoxels_to_nodes.to_dict())

    return supervoxel_historical_l2_ids


# TODO: I think this is another path to the above where you only have to check n_edits
# timepoints. Seemed like it could be faster but only by ~1/2
# stop_layer = 2
# # for all supervoxels, look up their level2 node now
# current_ts = client.timestamp
# node_ids = client.chunkedgraph.get_roots(
#     supervoxel_ids, stop_layer=stop_layer, timestamp=current_ts
# )
# node_to_supervoxel = pd.Series(index=node_ids, data=supervoxel_ids)
# node_to_supervoxel.index.name = "node_id"
# node_to_supervoxel.name = "supervoxel_id"

# # find out when they were made
# timestamps = client.chunkedgraph.get_root_timestamps(node_ids)
# node_timestamps = pd.Series(data=timestamps, index=node_ids)
# node_timestamps.index.name = "node_id"
# node_timestamps.name = "timestamp"

# # anything that was created at before oldest_ts we can ignore
# oldest_ts = client.chunkedgraph.get_oldest_timestamp()

# # everything else, we need to look up its history
# new_nodes = node_timestamps[node_timestamps > oldest_ts].index
# supervoxels_to_lookup = node_to_supervoxel.loc[new_nodes]

# from datetime import datetime
# from datetime import timedelta

# TIMESTAMP_DELTA = timedelta(microseconds=1)

# possible_timestamps = change_log["timestamp"]
# possible_timestamps = [datetime.fromisoformat(ts) for ts in possible_timestamps.values]

# # outs = []
# # for ts in tqdm(possible_timestamps[:10]):
# #     out = client.chunkedgraph.get_roots(
# #         supervoxels_to_lookup, stop_layer=2, timestamp=ts - TIMESTAMP_DELTA
# #     )
# #     outs.append(out)

# from joblib import Parallel, delayed
# from tqdm_joblib import tqdm_joblib

# with tqdm_joblib(total=len(possible_timestamps)):
#     outs = Parallel(n_jobs=-1)(
#         delayed(client.chunkedgraph.get_roots)(
#             supervoxels_to_lookup, stop_layer=2, timestamp=ts - TIMESTAMP_DELTA
#         )
#         for ts in possible_timestamps
#     )

# TODO: I think this is another alternative path but still have to lookup supervoxels
# def get_level2_lineage_components(networkdeltas_by_operation):
#     graph = nx.DiGraph()

#     for operation_id, delta in networkdeltas_by_operation.items():
#         for node1 in delta.removed_nodes:
#             for node2 in delta.added_nodes:
#                 graph.add_edge(node1, node2, operation_id=operation_id)

#     level2_lineage_components = list(nx.weakly_connected_components(graph))

#     level2_lineage_component_map = {}
#     for i, component in enumerate(level2_lineage_components):
#         for node in component:
#             level2_lineage_component_map[node] = i

#     level2_lineage_component_map = pd.Series(level2_lineage_component_map)

#     return level2_lineage_component_map


# level2_id_components = get_level2_lineage_components(networkdeltas)
# level2_id_components


def get_used_node_ids(initial_graph, edits, anchor_nodes):
    """Starting from an initial graph and a series of edits, get the nodes that are
    used in at least one state of the graph throughout its history.

    Parameters
    ----------
    initial_graph : nx.Graph
        The initial graph to start from.
    edits : dict
        A dictionary of edits where the key is the `operation_id` and the value is a
        `NetworkDelta` object.
    anchor_nodes : list
        A list of nodes that are on the object of interest, used to pick the connected
        component to consider at each point in the history.

    Returns
    -------
    :
        Nodes that are ever used in the history of the graph.
    """
    graph = initial_graph.copy()

    components = []
    subgraphs = []

    # remember to include the initial state
    edits = {-1: None, **edits}

    # after each edit, apply it and store the connected component for the nucleus node
    for edit_id, delta in edits.items():
        component = resolve_edit(graph, delta, anchor_nodes)
        components.append(component)

        subgraph = graph.subgraph(component).copy()
        subgraphs.append(subgraph)

    return np.unique(np.concatenate([list(c) for c in components]))


def get_changed_nodes(edits):
    """From a set of edits, get the nodes that have changed (added or removed)."""
    changed_nodes = []
    for _, edit in edits.items():
        if edit is not None:
            changed_nodes.extend(edit.added_nodes)
            changed_nodes.extend(edit.removed_nodes)
    return np.unique(changed_nodes)


def get_supervoxel_mappings(supervoxel_ids, edits, client, n_jobs=-1):
    """For a set of supervoxels and edits, get a mapping between the supervoxels and
    any level2 nodes they could have been part of across time."""
    # from our set of edits, get and level2 nodes that might have changed
    changed_nodes = get_changed_nodes(edits)

    # now, for all of those, look up supervoxels and see if they have any of the
    # supervoxels of interest
    def check_leaves(node):
        supervoxels = client.chunkedgraph.get_leaves(node)
        mask = np.isin(supervoxels, supervoxel_ids)
        return supervoxels[mask]

    with tqdm_joblib(desc="Getting leaves", total=len(changed_nodes)):
        leaves_by_l2 = Parallel(n_jobs=n_jobs)(
            delayed(check_leaves)(node) for node in changed_nodes
        )

    # store the mapping of supervoxels to level2 nodes that we got from looking at what
    # changed
    supervoxel_mappings = pd.Series(
        data=[[] for _ in range(len(supervoxel_ids))],
        index=supervoxel_ids,
        dtype=object,
    )
    for level2_id, leaf_supervoxel_ids in zip(changed_nodes, leaves_by_l2):
        for supervoxel_id in leaf_supervoxel_ids:
            supervoxel_mappings[supervoxel_id].append(level2_id)

    # anything without a mapping must not have changed
    unchanged_supervoxels = supervoxel_mappings[
        supervoxel_mappings.apply(len) == 0
    ].index

    # for those, simply look up their level2 nodes now
    chunk_size = 10000
    supervoxel_chunks = np.array_split(
        unchanged_supervoxels, len(unchanged_supervoxels) // chunk_size
    )
    for supervoxel_chunk in tqdm(supervoxel_chunks, desc="Getting remaining level2s"):
        level2_ids = client.chunkedgraph.get_roots(supervoxel_chunk, stop_layer=2)
        for supervoxel_id, level2_id in zip(supervoxel_chunk, level2_ids):
            supervoxel_mappings[supervoxel_id].append(level2_id)

    return supervoxel_mappings


def get_nucleus_location(root_id, client):
    nuc_table = client.info.get_datastack_info()["soma_table"]
    nuc_info = client.materialize.query_table(
        nuc_table,
        filter_equal_dict=dict(pt_root_id=root_id),
        desired_resolution=[1, 1, 1],
    )
    nuc_loc = nuc_info["pt_position"].values[0]
    return nuc_loc



from .graph_edits import (  # noqa: I001
    compare_graphs,
    get_detailed_change_log,
    get_metadata_table,
    get_metaedits,
    get_operation_level2_edit,
    get_operations_level2_edits,
    get_root_level2_edits,
    get_metaedit_counts,
    check_graph_changes,
)
from .level2_graph import get_initial_graph, get_level2_data, get_level2_spatial_graphs
from .networkdelta import NetworkDelta
from .utils import (
    get_node_aliases,
    get_component_masks,
    get_nucleus_supervoxel,
    get_nodes_aliases,
    get_supervoxel_mappings,
    get_changed_nodes,
    get_used_node_ids,
)
from .replay import apply_edit, resolve_edit, find_anchor_node, apply_edit_sequence
from .synapses import get_mutable_synapses, map_synapses_to_sequence
from .skeletons import skeletonize_sequence, compare_skeletons, check_skeleton_changes

__all__ = [
    "compare_graphs",
    "get_detailed_change_log",
    "get_metadata_table",
    "get_metaedits",
    "get_operation_level2_edit",
    "get_operations_level2_edits",
    "get_root_level2_edits",
    "get_initial_graph",
    "apply_edit",
    "NetworkDelta",
    "get_node_aliases",
    "get_component_masks",
    "get_initial_network",
    "get_nucleus_supervoxel",
    "get_mutable_synapses",
    "get_nodes_aliases",
    "resolve_edit",
    "get_used_node_ids",
    "get_supervoxel_mappings",
    "find_anchor_node",
    "get_changed_nodes",
    "get_used_node_ids",
    "apply_edit_sequence",
    "map_synapses_to_sequence",
    "skeletonize_sequence",
    "compare_skeletons",
    "check_skeleton_changes",
    "get_metaedit_counts",
    "check_graph_changes",
    "get_level2_data",
    "get_level2_spatial_graphs",
]



import time

import networkx as nx
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from .utils import get_supervoxel_mappings


def get_mutable_synapses(
    root_id,
    edits,
    client,
    sides="both",
    synapse_table=None,
    remove_self=True,
    verbose=False,
    n_jobs=-1,
):
    """Get all synapses that could have been part of this `root_id` across all states."""
    # TODO is it worth parallelizing this function?

    # TODO could also be sped up by taking the union of L2 IDS that get used, then
    # doing a current get_roots on those, feeding that into synapse query

    if synapse_table is None:
        synapse_table = client.info.get_datastack_info()["synapse_table"]

    if sides == "both":
        sides = ["pre", "post"]
    elif sides == "pre":
        sides = ["pre"]
    elif sides == "post":
        sides = ["post"]

    # find all of the original objects that at some point were part of this neuron
    t = time.time()
    original_roots = client.chunkedgraph.get_original_roots(root_id)
    if verbose:
        print(f"Getting original roots took {time.time() - t:.2f} seconds")

    # now get all of the latest versions of those objects
    # this will likely be a larger set of objects than we started with since those
    # objects could have seen further editing, etc.
    t = time.time()
    latest_roots = client.chunkedgraph.get_latest_roots(original_roots)
    if verbose:
        print(f"Getting latest roots took {time.time() - t:.2f} seconds")

    tables = []
    for side in sides:
        if verbose:
            print(f"Querying synapse table for {side}-synapses...")

        # get the pre/post-synapses that correspond to those objects
        t = time.time()
        syn_df: pd.DataFrame = client.materialize.query_table(
            synapse_table,
            filter_in_dict={f"{side}_pt_root_id": latest_roots},
        )
        syn_df.set_index("id", inplace=True)
        if verbose:
            print(f"Querying synapse table took {time.time() - t:.2f} seconds")

        if remove_self:
            syn_df.query("pre_pt_root_id != post_pt_root_id", inplace=True)

        tables.append(syn_df)

    all_supervoxel_ids = []
    for i, side in enumerate(sides):
        table = tables[i]
        all_supervoxel_ids.append(table[f"{side}_pt_supervoxel_id"].unique())
    supervoxel_ids = np.unique(np.concatenate(all_supervoxel_ids))
    supervoxel_mappings = get_supervoxel_mappings(
        supervoxel_ids, edits, client, n_jobs=n_jobs
    )

    exploded_tables = []
    for i, side in enumerate(sides):
        table = tables[i]
        table[f"{side}_pt_level2_id"] = table[f"{side}_pt_supervoxel_id"].map(
            supervoxel_mappings
        )
        exploded_tables.append(table.explode(f"{side}_pt_level2_id"))

    if len(exploded_tables) == 1:
        return exploded_tables[0]
    else:
        return exploded_tables[0], exploded_tables[1]


def map_synapses_to_sequence(
    synapses: pd.DataFrame,
    nodes_by_state: dict[list, nx.Graph],
    side="pre",
    verbose=True,
) -> dict[int, dict[int, int]]:
    """Map synapses (with level2 node information) to a sequence of level2 nodes/graphs.

    Parameters
    ----------
    synapses : pd.DataFrame
        A dataframe of synapses with a column "{side}_pt_level2_id" that describes what
        level2 node the synapse is on. Note that a single synapse ID can be associated
        with multiple level2 nodes, denoted by multiple rows. This specific synapse
        table can be generated using `paleo.get_mutable_synapses`.
    nodes_by_state : dict[list, nx.Graph]
        A dictionary mapping each state IDs to either a list of level2 nodes or a level2
        graph.
    side : str, optional
        The side of the synapse to map to the sequence of states, either "pre" or
        "post".
    verbose : bool, optional
        Whether to display a progress bar.

    Returns
    :
        Dictionary mapping each state IDs to a dictionary, where the keys are the
        synapse IDs and the values are the level2 node IDs they are associated with at
        that state.
    """
    if f"{side}_pt_level2_id" not in synapses.columns:
        raise ValueError(
            f"The synapses dataframe must have a column '{side}_pt_level2_id' to map synapses to components."
        )
    synapses = synapses.reset_index(drop=False).set_index(f"{side}_pt_level2_id")
    synapse_ids_by_edit = {}
    for state_id, nodes in tqdm(
        nodes_by_state.items(), desc="Mapping synapses to states", disable=not verbose
    ):
        component_synapse_index = synapses.index.intersection(list(nodes))
        # synapse_ids_at_state = (
        #     synapses.loc[component_synapse_index, "id"].unique().tolist()
        # )

        # synapse_ids_by_edit[state_id] = synapse_ids_at_state

        synapse_mapping_at_state = (
            synapses.loc[component_synapse_index, "id"]
            .to_frame()
            .reset_index()
            .set_index("id")[f"{side}_pt_level2_id"]
            .to_dict()
        )

        synapse_ids_by_edit[state_id] = synapse_mapping_at_state

    return synapse_ids_by_edit




from typing import Optional, Union

import networkx as nx
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from .networkdelta import NetworkDelta


def apply_edit(graph: nx.Graph, networkdelta: NetworkDelta):
    """Apply the edit described by the networkdelta to the graph."""
    removed_edges = networkdelta.removed_edges
    removed_nodes = networkdelta.removed_nodes

    added_edges = networkdelta.added_edges
    added_nodes = networkdelta.added_nodes

    graph.add_nodes_from(added_nodes)
    graph.add_edges_from(added_edges)

    # NOTE: these do not error if the nodes or edges are not in the graph
    # may want to revisit that
    graph.remove_nodes_from(removed_nodes)
    graph.remove_edges_from(removed_edges)


def find_anchor_node(graph, anchor_nodes):
    """Find the first anchor node that is in the graph."""
    for anchor_node in anchor_nodes:
        if graph.has_node(anchor_node):
            return anchor_node
    return None


def resolve_edit(
    graph: nx.Graph,
    networkdelta: Optional[NetworkDelta],
    anchor_nodes: list,
):
    """Apply the edit described by the networkdelta and return the connected component
    containing the anchor node."""
    if networkdelta is not None:
        apply_edit(graph, networkdelta)
    anchor_node = find_anchor_node(graph, anchor_nodes)
    component = nx.node_connected_component(graph, anchor_node)
    return component


def apply_edit_sequence(
    graph: nx.Graph,
    edits: dict,
    anchor_nodes: Union[list, pd.Index, np.ndarray, pd.Series],
    return_graphs: bool = False,
    include_initial: bool = True,
    remove_unchanged: bool = False,
    verbose: bool = True,
) -> Union[dict, tuple[dict, dict]]:
    """Apply a sequence of edits to the graph in order, storing information about
    intermediate states."""
    graph = graph.copy()
    if include_initial and -1 not in edits:
        edits = {-1: None, **edits}

    out = {}
    for edit_id, edit in tqdm(
        edits.items(), disable=not verbose, desc="Applying edits"
    ):
        component = resolve_edit(graph, edit, anchor_nodes)
        if return_graphs:
            out[edit_id] = graph.subgraph(component).copy()
        else:
            out[edit_id] = component

    if remove_unchanged:
        last = list(out.values())[0]
        for edit_id, current in list(out.items())[1:]:
            if return_graphs:
                if nx.utils.graphs_equal(last, current):
                    del out[edit_id]
                else:
                    last = current
            else:
                if set(last) == set(current):
                    del out[edit_id]
                else:
                    last = current

    return out



import json
import pprint
from typing import Collection

import numpy as np


class NetworkDelta:
    def __init__(
        self,
        removed_nodes: np.ndarray,
        added_nodes: np.ndarray,
        removed_edges: np.ndarray,
        added_edges: np.ndarray,
        metadata: dict = {},
    ):
        """
        A class to represent a change to a network.

        Parameters
        ----------
        removed_nodes :
            IDs of nodes that were removed by this operation.
        added_nodes :
            IDs of nodes that were added by this operation.
        removed_edges :
            Edges that were removed by this operation.
        added_edges :
            Edges that were added by this operation.
        metadata :
            A dictionary of arbitrary metadata about the operation.
        """
        self.removed_nodes = removed_nodes
        self.added_nodes = added_nodes
        self.removed_edges = removed_edges
        self.added_edges = added_edges
        self.metadata = metadata

    def __repr__(self):
        rep = "NetworkDelta(\n"
        rep += f"   removed_nodes: {self.removed_nodes.shape[0]},\n"
        rep += f"   added_nodes: {self.added_nodes.shape[0]},\n"
        rep += f"   removed_edges: {self.removed_edges.shape[0]},\n"
        rep += f"   added_edges: {self.added_edges.shape[0]},\n"
        if len(self.metadata) > 0:
            rep += "   metadata: {\n"
            rep += " " + pprint.pformat(self.metadata, indent=6)[1:-1]
            rep += "\n   }\n"
            rep += ")"
        else:
            rep += "   metadata: {}\n"
            rep += ")"
        return rep

    def to_dict(self) -> dict:
        out = dict(
            removed_nodes=self.removed_nodes.tolist(),
            added_nodes=self.added_nodes.tolist(),
            removed_edges=self.removed_edges.tolist(),
            added_edges=self.added_edges.tolist(),
            metadata=self.metadata,
        )
        return out

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, input, dtype=int):
        removed_nodes = np.array(input["removed_nodes"], dtype=dtype)
        added_nodes = np.array(input["added_nodes"], dtype=dtype)
        removed_edges = np.atleast_2d(np.array(input["removed_edges"], dtype=dtype))
        # TODO these size things are a hack, will not generalize if nodes are 3D
        if removed_edges.size == 0:
            removed_edges = np.empty((0, 2), dtype=dtype)
        added_edges = np.atleast_2d(np.array(input["added_edges"], dtype=dtype))
        if added_edges.size == 0:
            added_edges = np.empty((0, 2), dtype=dtype)
        metadata = input["metadata"]
        return cls(
            removed_nodes, added_nodes, removed_edges, added_edges, metadata=metadata
        )

    # @classmethod
    # def from_dict(cls, input):
    #     removed_nodes = pd.DataFrame(index=input["removed_nodes"])
    #     added_nodes = pd.DataFrame(index=input["added_nodes"])
    #     removed_edges = pd.DataFrame(
    #         input["removed_edges"], columns=["source", "target"]
    #     )
    #     added_edges = pd.DataFrame(input["added_edges"], columns=["source", "target"])
    #     metadata = input["metadata"]
    #     return cls(
    #         removed_nodes, added_nodes, removed_edges, added_edges, metadata=metadata
    #     )

    # @classmethod
    # def from_json(cls, input):
    #     return cls.from_dict(json.loads(input))

    def __eq__(self, other: "NetworkDelta") -> bool:
        if not isinstance(other, NetworkDelta):
            return False
        if not np.array_equal(self.removed_nodes, other.removed_nodes):
            return False
        if not np.array_equal(self.added_nodes, other.added_nodes):
            return False
        if not np.array_equal(self.removed_edges, other.removed_edges):
            return False
        if not np.array_equal(self.added_edges, other.added_edges):
            return False
        return True

    def __ne__(self, other: "NetworkDelta") -> bool:
        return not self.__eq__(other)

    def __add__(self, other: "NetworkDelta") -> "NetworkDelta":
        return combine_deltas([self, other])

    @property
    def is_empty(self) -> bool:
        return (
            self.removed_nodes.size == 0
            and self.added_nodes.size == 0
            and self.removed_edges.size == 0
            and self.added_edges.size == 0
        )


def _unique_concatenate(
    arrays: Collection[np.ndarray], verify_integrity=True
) -> np.ndarray:
    concat_array = np.concatenate(arrays)
    if verify_integrity:
        unique_array, unique_counts = np.unique(
            concat_array, axis=0, return_counts=True
        )
        if np.any(unique_counts > 1):
            raise ValueError("Duplicate values found in arrays")
    return concat_array


def combine_deltas(deltas: Collection[NetworkDelta]) -> NetworkDelta:
    total_added_nodes = _unique_concatenate([delta.added_nodes for delta in deltas])
    total_removed_nodes = _unique_concatenate([delta.removed_nodes for delta in deltas])
    total_added_edges = _unique_concatenate([delta.added_edges for delta in deltas])
    total_removed_edges = _unique_concatenate([delta.removed_edges for delta in deltas])

    return NetworkDelta(
        total_removed_nodes,
        total_added_nodes,
        total_removed_edges,
        total_added_edges,
    )



import json

import networkx as nx
import numpy as np

from paleo import NetworkDelta


def edits_to_json(networkdeltas_by_operation: dict) -> str:
    networkdelta_dicts = {}
    for operation_id, delta in networkdeltas_by_operation.items():
        networkdelta_dicts[operation_id] = delta.to_dict()
    return json.dumps(networkdelta_dicts)


def json_to_edits(edits_json) -> dict:
    networkdelta_dicts = json.loads(edits_json)
    networkdeltas_by_operation = {}
    for operation_id, delta in networkdelta_dicts.items():
        networkdeltas_by_operation[int(operation_id)] = NetworkDelta.from_dict(delta)
    return networkdeltas_by_operation


def graph_to_json(graph: nx.Graph):
    edges = np.array(list(graph.edges(data=False))).astype(str).tolist()
    nodes = np.array(list(graph.nodes(data=False))).astype(str).tolist()

    return json.dumps({"edges": edges, "nodes": nodes})


def json_to_graph(json_str):
    data = json.loads(json_str)
    nodes = np.array(data["nodes"]).astype(int)
    edges = np.array(data["edges"]).astype(int)
    graph = nx.Graph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph



from typing import Literal, Optional

import networkx as nx
import numpy as np
import pandas as pd
from caveclient import CAVEclient
from joblib import Parallel, delayed
from tqdm.auto import tqdm
from tqdm_joblib import tqdm_joblib

from .graph_edits import compare_graphs
from .networkdelta import NetworkDelta
from .utils import get_nucleus_location


def skeletonize_sequence(
    graphs_by_state: dict,
    client: Optional[CAVEclient] = None,
    root_id: Optional[int] = None,
    root_point: Optional[np.ndarray] = None,
    level2_data: Optional[pd.DataFrame] = None,
    remove_unchanged: bool = False,
    return_as: Literal["meshparty", "arrays", "networkx"] = "meshparty",
    n_jobs: int = -1,
    verbose: bool = True,
):
    """Generate skeletons for a sequence of graphs.

    Parameters
    ----------
    graphs_by_state :
        A dictionary mapping state IDs to NetworkX graphs.
    client :
        A CAVEclient instance.
    root_id :
        The ID of the root node, used to determine the root point for skeletonization if
        `root_point` is not provided.
    root_point :
        The root point for skeletonization. If not provided, it will be determined
        using the `root_id` and `client`.
    level2_data :
        The level 2 data table, containing the node coordinates in columns
        "rep_coord_nm_x", "rep_coord_nm_y", "rep_coord_nm_z".
    remove_unchanged :
        Whether to remove states with no changes from the previous state from the
        sequence.
    return_as :
        The format to return the skeletons in. Options are "meshparty", "arrays", and
        "networkx".
    n_jobs :
        The number of jobs to run in parallel. See joblib.Parallel for more details.
    verbose :
        Whether to display progress bars.
    """

    try:
        from pcg_skel import pcg_skeleton_direct
    except (ImportError, ModuleNotFoundError):
        msg = (
            "Please install the `pcg_skel` package to use skeletonization features. "
            "This can be done by running `pip install pcg-skel` "
            "or `pip install paleo[skeleton]`."
        )
        raise ModuleNotFoundError(msg)

    if level2_data is None:
        used_nodes = set()
        for graph in graphs_by_state.values():
            used_nodes.update(graph.nodes())
        # TODO add code for getting it using existing tools
        used_nodes = np.array(list(used_nodes))
        level2_data = client.l2cache.get_l2data_table(used_nodes)

    if root_point is None:
        root_point = get_nucleus_location(root_id, client)

    def _skeletonize_state(graph):
        node_ids = pd.Index(list(graph.nodes()))
        vertices = level2_data.loc[
            node_ids, ["rep_coord_nm_x", "rep_coord_nm_y", "rep_coord_nm_z"]
        ].values
        edges = nx.to_pandas_edgelist(graph).values
        edges = np.vectorize(node_ids.get_loc)(edges)

        skeleton = pcg_skeleton_direct(
            vertices, edges, root_point=root_point, collapse_soma=True
        )

        mapping = dict(zip(node_ids, skeleton.mesh_to_skel_map.tolist()))
        return skeleton, mapping

    if n_jobs != 1:
        with tqdm_joblib(
            desc="Skeletonizing states", disable=not verbose, total=len(graphs_by_state)
        ):
            results = Parallel(n_jobs=n_jobs)(
                delayed(_skeletonize_state)(graph) for graph in graphs_by_state.values()
            )
    else:
        results = [
            _skeletonize_state(graph)
            for graph in tqdm(
                graphs_by_state.values(),
                desc="Skeletonizing states",
                disable=not verbose,
                total=len(graphs_by_state),
            )
        ]

    if remove_unchanged:
        keep_state = check_skeleton_changes(
            {
                state_id: result[0]
                for state_id, result in zip(graphs_by_state.keys(), results)
            }
        )

    skeletons_by_state = {}
    mappings_by_state = {}
    for state_id, (skeleton, mapping) in zip(graphs_by_state.keys(), results):
        if remove_unchanged and not keep_state[state_id]:
            continue
        if return_as == "meshparty":
            pass
        elif return_as == "arrays":
            skeleton = (skeleton.vertices, skeleton.edges)
        elif return_as == "networkx":
            if len(skeleton.edges) > 0:
                new_skeleton = nx.from_edgelist(skeleton.edges.tolist())
            else:
                new_skeleton = nx.Graph()
                new_skeleton.add_nodes_from(list(range(len(skeleton.vertices))))
            positions = {}
            for i in range(len(skeleton.vertices)):
                positions[i] = {}
                positions[i]["x"] = skeleton.vertices[i][0].item()
                positions[i]["y"] = skeleton.vertices[i][1].item()
                positions[i]["z"] = skeleton.vertices[i][2].item()
            nx.set_node_attributes(new_skeleton, positions)
            skeleton = new_skeleton
        skeletons_by_state[state_id] = skeleton
        mappings_by_state[state_id] = mapping
    return skeletons_by_state, mappings_by_state


def compare_skeletons(
    skeleton1, skeleton2, return_as_spatial=False
) -> tuple[NetworkDelta, np.ndarray]:
    """Compare two skeletons, finding the network changes between them.

    Parameters
    ----------
    skeleton1 :
        The first skeleton to compare.
    skeleton2 :
        The second skeleton to compare.
    return_as_spatial : bool, optional
        Whether to return the changes in the original spatial coordinates, by default
        False.

    Returns
    -------
    :
        The changes between the two skeletons.
    :
        Array of the positions of all nodes referenced between the two skeletons. Note
        that the deltas are references to the positions in this array, unless they were
        already replaced if `return_as_spatial=True`.
    """
    # get positions of all of the nodes referenced between the two skeletons
    vertices1 = skeleton1.vertices
    index1 = pd.MultiIndex.from_arrays(vertices1.T)
    vertices2 = skeleton2.vertices
    index2 = pd.MultiIndex.from_arrays(vertices2.T)
    union_index: pd.MultiIndex = index1.union(index2)

    # reindex edges to be in this common reference frame
    vertices1_to_ids = union_index.get_indexer_for(index1)
    vertices2_to_ids = union_index.get_indexer_for(index2)
    edges1 = skeleton1.edges
    remapped_edges1 = np.vectorize(vertices1_to_ids.__getitem__, otypes=["int"])(edges1)
    edges2 = skeleton2.edges
    remapped_edges2 = np.vectorize(vertices2_to_ids.__getitem__, otypes=["int"])(edges2)

    # now can use the vanilla comparison function
    delta = compare_graphs(remapped_edges1, remapped_edges2)

    union_positions = union_index.to_frame().values

    # optionally, put things back in the original spatial coordinates
    # otherwise, the user can use union_index_array to do this themselves
    if return_as_spatial:
        delta.removed_nodes = union_positions[delta.removed_nodes]
        delta.added_nodes = union_positions[delta.added_nodes]
        delta.removed_edges = union_positions[delta.removed_edges]
        delta.added_edges = union_positions[delta.added_edges]

    return delta, union_positions


def check_skeleton_changes(skeletons_by_state: dict) -> dict:
    """Check if each skeleton in a sequence is different,
    compared to the one before it.

    Parameters
    ----------
    skeletons_by_state : dict
        A dictionary mapping state IDs to skeletons.

    Returns
    -------
    :
        A dictionary mapping state IDs to whether the skeleton is different from the
        previous one.
    """
    states = list(skeletons_by_state.keys())
    states_is_new = {states[0]: True}
    deltas = []
    for state1, state2 in zip(states, states[1:]):
        skeleton1 = skeletons_by_state[state1]
        skeleton2 = skeletons_by_state[state2]
        delta, _ = compare_skeletons(skeleton1, skeleton2)
        if not delta.is_empty:
            deltas.append(delta)
            states_is_new[state2] = True
        else:
            states_is_new[state2] = False
    return states_is_new



import warnings

import networkx as nx
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import TqdmExperimentalWarning

warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)
from typing import Optional

from caveclient import CAVEclient
from tqdm.auto import tqdm
from tqdm_joblib import tqdm_joblib

from .utils import _get_level2_nodes_edges, _sort_edgelist


def get_initial_node_ids(root_id, client):
    lineage_g = client.chunkedgraph.get_lineage_graph(root_id, as_nx_graph=True)
    node_in_degree = pd.Series(dict(lineage_g.in_degree()))
    original_node_ids = node_in_degree[node_in_degree == 0].index
    return original_node_ids


def get_initial_graph(root_id, client, verbose=True, return_as="networkx", n_jobs=-1):
    """Get the initial graph for a given `root_id`, including objects that could become
    part of the neuron in the future."""
    if return_as not in ["networkx", "arrays"]:
        ValueError(f"`return_as` must be 'networkx' or 'arrays', got {return_as}")

    original_node_ids = get_initial_node_ids(root_id, client)

    def _get_info_for_node(leaf_id):
        nodes, edges = _get_level2_nodes_edges(leaf_id, client)
        return nodes, edges

    if n_jobs != 1:
        with tqdm_joblib(
            total=len(original_node_ids),
            disable=not verbose,
            desc="Getting initial graph",
        ):
            outs = Parallel(n_jobs=n_jobs)(
                delayed(_get_info_for_node)(leaf_id) for leaf_id in original_node_ids
            )
    else:
        outs = []
        for leaf_id in tqdm(original_node_ids):
            outs.append(_get_info_for_node(leaf_id))

    all_nodes = []
    all_edges = []
    for out in outs:
        nodes, edges = out
        all_nodes.append(nodes)
        all_edges.append(edges)

    all_nodes = np.concatenate(all_nodes, axis=0)
    all_edges = np.concatenate(all_edges, axis=0)

    all_nodes = np.unique(all_nodes)
    all_edges = _sort_edgelist(all_edges)

    if return_as == "networkx":
        graph = nx.Graph()
        graph.add_nodes_from(all_nodes)
        graph.add_edges_from(all_edges)
        return graph
    else:  # return_as == 'arrays'
        return all_nodes, all_edges


def get_level2_data(graphs_by_state: dict, client: CAVEclient):
    used_nodes = set()
    for graph in graphs_by_state.values():
        used_nodes.update(graph.nodes())
    used_nodes = np.array(list(used_nodes))
    level2_data = client.l2cache.get_l2data_table(used_nodes)
    return level2_data


def get_level2_spatial_graphs(
    graphs_by_state: dict,
    client: Optional[CAVEclient] = None,
    level2_data=None,
    index_on=None,
):
    if level2_data is None and client is not None:
        level2_data = get_level2_data(graphs_by_state, client)
    elif level2_data is None:
        raise ValueError("Must provide either `level2_data` or `client`")
    
    if index_on is None:
        index_on = [
            "rep_coord_nm_x",
            "rep_coord_nm_y",
            "rep_coord_nm_z",
            "size_nm3",
            "area_nm2",
        ]
    keys = level2_data[index_on]
    keys = list(zip(*keys.values.T))
    key_mapping = dict(zip(level2_data.index, keys))

    spatial_graphs_by_state = {}
    for state_id, graph in graphs_by_state.items():
        spatial_graph = graph.copy()
        spatial_graph = nx.relabel_nodes(spatial_graph, key_mapping)
        spatial_graphs_by_state[state_id] = spatial_graph

    return spatial_graphs_by_state



import warnings
from datetime import datetime
from typing import Collection, Optional, Union

import networkx as nx
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy.sparse import csr_array
from scipy.sparse.csgraph import connected_components
from tqdm import TqdmExperimentalWarning

warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)

from caveclient import CAVEclient
from tqdm.auto import tqdm
from tqdm_joblib import tqdm_joblib

from .constants import TIMESTAMP_DELTA
from .networkdelta import NetworkDelta, combine_deltas
from .types import Graph, Integer, Number
from .utils import _get_level2_nodes_edges, _sort_edgelist


def _get_changed_edges(
    before_edges: np.ndarray, after_edges: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    before_edges = _sort_edgelist(before_edges)
    after_edges = _sort_edgelist(after_edges)

    before_edges = np.concatenate(
        (before_edges, np.zeros((before_edges.shape[0], 1), dtype=int)), axis=1
    )
    after_edges = np.concatenate(
        (after_edges, np.ones((after_edges.shape[0], 1), dtype=int)), axis=1
    )

    all_edges = np.concatenate((before_edges, after_edges), axis=0, dtype=int)
    unique_edges, index, edge_counts = np.unique(
        all_edges[:, :2], axis=0, return_counts=True, return_index=True
    )

    single_inverse = index[edge_counts == 1]
    single_edges = all_edges[single_inverse]

    removed_edges = single_edges[single_edges[:, 2] == 0][:, :2]
    added_edges = single_edges[single_edges[:, 2] == 1][:, :2]

    return removed_edges, added_edges


def _make_bbox(
    bbox_radius: Number, point_in_seg: np.ndarray, seg_resolution: np.ndarray
) -> np.ndarray:
    point_in_nm = point_in_seg * seg_resolution
    x_center, y_center, z_center = point_in_nm

    x_start = x_center - bbox_radius
    x_stop = x_center + bbox_radius
    y_start = y_center - bbox_radius
    y_stop = y_center + bbox_radius
    z_start = z_center - bbox_radius
    z_stop = z_center + bbox_radius

    start_point_cg = np.array([x_start, y_start, z_start]) / seg_resolution
    stop_point_cg = np.array([x_stop, y_stop, z_stop]) / seg_resolution

    bbox_cg = np.array([start_point_cg, stop_point_cg], dtype=int)
    return bbox_cg


def _get_all_nodes_edges(
    root_ids: Number, client: CAVEclient, bounds: Optional[np.ndarray] = None
) -> tuple[np.ndarray, np.ndarray]:
    all_nodes = []
    all_edges = []
    for root_id in root_ids:
        nodes, edges = _get_level2_nodes_edges(root_id, client, bounds=bounds)
        all_nodes.append(nodes)
        all_edges.append(edges)
    if len(all_nodes) == 0:
        return np.empty(0, dtype=int), np.empty((0, 2), dtype=int)
    else:
        all_nodes = np.concatenate(all_nodes, dtype=int)
        all_edges = np.concatenate(all_edges, dtype=int)
        return all_nodes, all_edges


def get_detailed_change_log(
    root_id: int, client: CAVEclient, filtered: bool = True
) -> pd.DataFrame:
    """Get a detailed change log for a root ID.

    Parameters
    ----------
    root_id :
        The root ID to get the change log for.
    client :
        The CAVEclient instance to use.
    filtered :
        Whether to filter the change log to only include changes which affect the
        final state of the root ID.

    Returns
    -------
    :
        A detailed change log for the root ID.
    """
    cg = client.chunkedgraph
    change_log = cg.get_tabular_change_log(root_id, filtered=filtered)[root_id]

    change_log.set_index("operation_id", inplace=True)
    change_log.sort_values("timestamp", inplace=True)
    change_log.drop(columns=["timestamp"], inplace=True)

    chunk_size = 500  # not sure exactly what the limit is here
    details = {}
    for i in range(0, len(change_log), chunk_size):
        sub_details = cg.get_operation_details(
            change_log.index[i : i + chunk_size].to_list()
        )
        details.update(sub_details)
    assert len(details) == len(change_log)

    details = pd.DataFrame(details).T
    details.index.name = "operation_id"
    details.index = details.index.astype(int)

    change_log = change_log.join(details)

    return change_log


def _get_nodes_edges_from_graph(graph):
    if isinstance(graph, tuple):
        nodes, edges = graph
    else:
        nodes = np.unique(graph.flatten())
        edges = graph
    return nodes, edges


def compare_graphs(
    graph_before: Graph, graph_after: Graph, metadata=False
) -> NetworkDelta:
    """Compare two graphs and return the differences.

    Parameters
    ----------
    graph_before :
        The graph before the operation. Can either be a tuple of (nodes, edges) stored
        as `np.ndarrays`, or just the edges as an `np.ndarray`.
    graph_after :
        The graph after the operation. Can either be a tuple of (nodes, edges) stored
        as `np.ndarrays`, or just the edges as an `np.ndarray`.

    Returns
    -------
    :
        The differences between the two graphs.
    """

    nodes_before, edges_before = _get_nodes_edges_from_graph(graph_before)
    nodes_after, edges_after = _get_nodes_edges_from_graph(graph_after)

    removed_nodes = np.setdiff1d(nodes_before, nodes_after)
    added_nodes = np.setdiff1d(nodes_after, nodes_before)

    removed_edges, added_edges = _get_changed_edges(edges_before, edges_after)

    # keep track of what changed
    if metadata:
        metadata_dict = {
            "n_added_nodes": len(added_nodes),
            "n_removed_nodes": len(removed_nodes),
            "n_modified_nodes": len(added_nodes) + len(removed_nodes),
            "n_added_edges": len(added_edges),
            "n_removed_edges": len(removed_edges),
            "n_modified_edges": len(added_edges) + len(removed_edges),
        }
    else:
        metadata_dict = {}

    return NetworkDelta(
        removed_nodes,
        added_nodes,
        removed_edges,
        added_edges,
        metadata=metadata_dict,
    )


def get_operation_level2_edit(
    operation_id: int,
    client: CAVEclient,
    before_root_ids: Optional[Collection[int]] = None,
    after_root_ids: Optional[Collection[int]] = None,
    timestamp: Optional[datetime] = None,
    point: Optional[np.ndarray] = None,
    radius: Number = 20_000,
    metadata: bool = False,
) -> NetworkDelta:
    """Extract changes to the level2 graph for a specific operation.

    Parameters
    ----------
    operation_id :
        The operation ID to extract changes for.
    client :
        The CAVEclient instance to use.
    before_root_ids :
        The root ID(s) that were involved in the operation prior to it happening. If
        None, these will be looked up.
    after_root_ids :
        The root ID(s) that were created by the operation. If None, these will be
        looked up.
    timestamp :
        The timestamp of the operation. Only used if `before_root_ids` is not provided.
        If None, this will be looked up.
    point :
        The point to center the bounding box on. If None, will compare the entire
        level2 graphs of the objects before and after the operation.
    radius :
        The radius of the bounding box to use.
    metadata :
        Whether to include metadata about the changes in the output.

    Returns
    -------
    :
        The changes to the level2 graph from this operation.
    """
    if before_root_ids is None and timestamp is not None:
        maps = client.chunkedgraph.get_past_ids(
            after_root_ids, timestamp_past=timestamp - TIMESTAMP_DELTA
        )
        before_root_ids = []
        for root in after_root_ids:
            before_root_ids.extend(maps["past_id_map"][root])

    # if the point to center on is not provided, or if there is no list of ids that
    # came before this edit, then we need to look them up
    if (
        (point is None and radius is not None)
        or (after_root_ids is None)
        or (before_root_ids is None)  # implies timestamp is None because of the above
    ):
        details = client.chunkedgraph.get_operation_details([operation_id])[
            str(operation_id)
        ]
        if point is None:
            point = np.array(details["sink_coords"][0])
        if after_root_ids is None:
            after_root_ids = details["roots"]
        if before_root_ids is None:
            timestamp = datetime.fromisoformat(details["timestamp"])
            pre_timestamp = timestamp - TIMESTAMP_DELTA
            maps = client.chunkedgraph.get_past_ids(
                after_root_ids, timestamp_past=pre_timestamp
            )
            before_root_ids = []
            for root in after_root_ids:
                before_root_ids.extend(maps["past_id_map"][root])

    if radius is None:
        bbox_cg = None
    else:
        bbox_cg = _make_bbox(radius, point, client.chunkedgraph.base_resolution).T

    # grabbing the union of before/after nodes/edges
    # NOTE: this is where all the compute time comes from
    all_before_nodes, all_before_edges = _get_all_nodes_edges(
        before_root_ids, client, bounds=bbox_cg
    )
    all_after_nodes, all_after_edges = _get_all_nodes_edges(
        after_root_ids, client, bounds=bbox_cg
    )

    networkdelta = compare_graphs(
        (all_before_nodes, all_before_edges), (all_after_nodes, all_after_edges)
    )
    if metadata:
        networkdelta.metadata["operation_id"] = operation_id

    return networkdelta


def get_operations_level2_edits(
    operation_ids: Union[Collection[Integer], Integer],
    client: CAVEclient,
    radius: Number = 20_000,
    metadata: bool = False,
    n_jobs: int = -1,
    verbose: bool = True,
) -> dict[Integer, NetworkDelta]:
    """Extract changes to the level2 graph for a list of operations.


    Parameters
    ----------
    operation_ids :
        The operation ID(s) to extract changes for.
    client :
        The CAVEclient instance to use.
    radius :
        The radius of the bounding box to use.
    metadata :
        Whether to include metadata about the changes in the output.
    n_jobs :
        The number of jobs to run in parallel. If -1, will use all available cores.
    verbose :
        Whether to display a progress bar.

    Returns
    -------
    :
        The changes to the level2 graph from these operations
    """
    if isinstance(operation_ids, (int, np.integer)):
        operation_ids = [operation_ids]
    if not isinstance(operation_ids, list):
        try:
            operation_ids = list(operation_ids)
        except TypeError:
            raise TypeError(
                f"`operation_ids` could not be coerced to a list: {operation_ids}"
            )

    details_by_operation = client.chunkedgraph.get_operation_details(operation_ids)
    new_roots_by_operation = {
        int(operation_id): details["roots"]
        for operation_id, details in details_by_operation.items()
    }
    timestamps_by_operation = {
        int(operation_id): datetime.fromisoformat(details["timestamp"])
        for operation_id, details in details_by_operation.items()
    }

    inputs_by_operation = []
    for operation_id in operation_ids:
        inputs_by_operation.append(
            {
                "operation_id": operation_id,
                "client": client,
                "before_root_ids": None,
                "after_root_ids": new_roots_by_operation[operation_id],
                "timestamp": timestamps_by_operation[operation_id],
                "radius": radius,
                "metadata": metadata,
            }
        )
    if n_jobs != 1:
        with tqdm_joblib(
            total=len(inputs_by_operation),
            disable=not verbose,
            desc="Extracting level2 edits",
        ):
            networkdeltas = Parallel(n_jobs=-1)(
                delayed(get_operation_level2_edit)(**inputs)
                for inputs in inputs_by_operation
            )
    else:
        networkdeltas = []
        for inputs in tqdm(
            inputs_by_operation,
            disable=not verbose,
            desc="Extracting level2 edits",
        ):
            networkdeltas.append(get_operation_level2_edit(**inputs))

    networkdeltas = dict(zip(operation_ids, networkdeltas))
    return networkdeltas


def get_root_level2_edits(
    root_id: Integer,
    client: CAVEclient,
    radius: Number = 20_000,
    metadata: bool = False,
    filtered: bool = False,
    n_jobs: int = -1,
    verbose: bool = True,
) -> dict[Integer, NetworkDelta]:
    """Extract changes to the level2 graph for all operations on a root.

    Parameters
    ----------
    root_id :
        The root ID to extract changes for.
    client :
        The CAVEclient instance to use.
    radius :
        The radius of the bounding box to use.
    metadata :
        Whether to include metadata about the changes in the output.
    filtered :
        Whether to filter the change log to only include changes which affect the
        final state of the root ID.
    n_jobs :
        The number of jobs to run in parallel. If -1, will use all available cores.
    verbose :
        Whether to display a progress bar.

    Returns
    -------
    :
        The changes to the level2 graph from each operation
    """
    change_log = get_detailed_change_log(root_id, client, filtered=filtered)

    inputs_by_operation = []
    for operation_id, row in change_log.iterrows():
        inputs_by_operation.append(
            {
                "operation_id": operation_id,
                "client": client,
                "before_root_ids": row["before_root_ids"],
                "after_root_ids": row["roots"],
                "timestamp": None,  # not needed since we know roots before/after
                "radius": radius,
                "metadata": metadata,
            }
        )

    if n_jobs != 1:
        with tqdm_joblib(
            total=len(inputs_by_operation),
            disable=not verbose,
            desc="Extracting level2 edits",
        ):
            networkdeltas_by_operation = Parallel(n_jobs=-1)(
                delayed(get_operation_level2_edit)(**inputs)
                for inputs in inputs_by_operation
            )
    else:
        networkdeltas_by_operation = []
        for inputs in tqdm(
            inputs_by_operation,
            disable=not verbose,
            desc="Extracting level2 edits",
        ):
            networkdeltas_by_operation.append(get_operation_level2_edit(**inputs))

    networkdeltas_by_operation = dict(zip(change_log.index, networkdeltas_by_operation))

    return networkdeltas_by_operation


def get_metaedits(
    networkdeltas: dict[Integer, NetworkDelta],
) -> tuple[dict[Integer, NetworkDelta], dict[Integer, list[Integer]]]:
    """Combine edits into meta-edits based on shared nodes.

    Meta-edits are groups of one or more edits which affected a local region in the
    chunkedgraph. More specifically, they are defined as groups of edits which are
    connected components in a graph where nodes are edits and edges are shared nodes
    between edits.

    Parameters
    ----------
    networkdeltas :
        The changes to the level2 graph from each operation.

    Returns
    -------
    :
        The changes to the level2 graph from each meta-operation.
    :
        A mapping of meta-operation IDs to the operation IDs that make them up.

    """
    # find the nodes that are modified in any way by each operation
    mod_sets = {}
    for edit_id, delta in networkdeltas.items():
        mod_set = []
        mod_set += delta.added_nodes.tolist()
        mod_set += delta.removed_nodes.tolist()
        try:
            mod_set += delta.added_edges[:, 0].tolist()
        except IndexError as e:
            print(delta.added_edges)
            print(type(delta.added_edges))
            print(delta.added_edges.shape)
            raise e
        mod_set += delta.added_edges[:, 1].tolist()
        mod_set += delta.removed_edges[:, 0].tolist()
        mod_set += delta.removed_edges[:, 1].tolist()
        mod_set = np.unique(mod_set)
        mod_sets[edit_id] = mod_set

    # make an incidence matrix of which nodes are modified by which operations
    index = np.unique(np.concatenate(list(mod_sets.values())))
    node_edit_indicators = pd.DataFrame(
        index=index, columns=networkdeltas.keys(), data=False
    )
    for edit_id, mod_set in mod_sets.items():
        node_edit_indicators.loc[mod_set, edit_id] = True

    # this inner product matrix tells us which operations are connected with at least
    # one overlapping node in common
    X = csr_array(node_edit_indicators.values.astype(int))
    product = X.T @ X

    # meta-operations are connected components in the above graph
    _, labels = connected_components(product, directed=False)

    meta_operation_map = {}
    operation_map = {}
    for label in np.unique(labels):
        edits = node_edit_indicators.columns[labels == label].tolist()
        meta_operation_map[label] = edits
        for edit in edits:
            operation_map[edit] = label.item()

    # for each meta-operation, combine the deltas of the operations that make it up
    networkdeltas_by_meta_operation = {}
    for meta_operation_id, operation_ids in meta_operation_map.items():
        meta_operation_id = int(meta_operation_id)
        deltas = [networkdeltas[operation_id] for operation_id in operation_ids]
        meta_networkdelta = combine_deltas(deltas)
        networkdeltas_by_meta_operation[meta_operation_id] = meta_networkdelta

    return networkdeltas_by_meta_operation, operation_map


def get_metaedit_counts(edit_map):
    metaedit_counts = {}
    for metaedit in edit_map.values():
        if metaedit not in metaedit_counts:
            metaedit_counts[metaedit] = 0
        metaedit_counts[metaedit] += 1
    return metaedit_counts


def get_metadata_table(operation_ids=None, root_ids=None, client=None):
    """Retrieve metadata for a list of operations or root IDs.

    NOTE: aspirational, not yet implemented. To make this efficient, would probably need
    a server-side implementation

    Parameters
    ----------
    operation_ids : list of int, optional
        The operation IDs to retrieve metadata for.
    root_ids : list of int, optional
        The root IDs to retrieve metadata for.
    client : CAVEclient, optional
        The CAVEclient instance to use.

    Returns
    -------
    pd.DataFrame
        The metadata for the operations or root IDs. Metadata includes:

        - `operation_id`: The operation ID.
        - `timestamp`: The timestamp of the operation.
        - `location`: The approximate [x,y,z] centroid of the operation in nanometers.
        - `volume_added`: The volume added by the operation in cubic nanometers.
        - `volume_removed`: The volume removed by the operation in cubic nanometers.
        - `n_added_nodes`: The number of level2 nodes added by the operation.
        - `n_removed_nodes`: The number of level2 nodes removed by the operation.
        - `n_modified_nodes`: The number of level2 nodes modified by the operation.
        - `n_added_edges`: The number of level2 edges added by the operation.
        - `n_removed_edges`: The number of level2 edges removed by the operation.
        - `n_modified_edges`: The number of level2 edges modified by the operation.

    """
    raise NotImplementedError("This function is not yet implemented.")


def check_graph_changes(graphs_by_state: dict):
    states = list(graphs_by_state.keys())
    states_is_new = {states[0]: True}
    for state_iloc in range(1, len(states)):
        last_state = states[state_iloc - 1]
        this_state = states[state_iloc]
        last_graph = graphs_by_state[last_state]
        this_graph = graphs_by_state[this_state]
        graphs_changed = not nx.utils.graphs_equal(last_graph, this_graph)
        states_is_new[this_state] = graphs_changed
    return states_is_new
